from meltano.core.plugin import PluginInstall, PluginType
from meltano.core.plugin_invoker import PluginInvoker


class DbtPlugin(PluginInstall):
    __plugin_type__ = PluginType.TRANSFORMERS

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)

    def invoker(self, session, project, *args, **kwargs):
        return DbtInvoker(
            session,
            project,
            self,
            *args,
            run_dir=project.root_dir("transform"),
            **kwargs
        )


class DbtTransformPlugin(PluginInstall):
    __plugin_type__ = PluginType.TRANSFORMS

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)


class DbtInvoker(PluginInvoker):
    def Popen_options(self):
        options = super().Popen_options()
        options["cwd"] = str(self.project.root_dir("transform"))

        return options



import logging
import requests
import threading
import time
import webbrowser
import psutil
from colorama import Fore

from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler, EVENT_TYPE_MODIFIED
from meltano.core.project import Project
from meltano.core.plugin import PluginInstall
from meltano.core.config_service import ConfigService
from meltano.core.compiler.project_compiler import ProjectCompiler
from meltano.core.plugin_invoker import invoker_factory
from meltano.api.models import db


class CompileEventHandler(PatternMatchingEventHandler):
    def __init__(self, compiler):
        self.compiler = compiler

        super().__init__(ignore_patterns=["*.m5oc"])

    def on_any_event(self, event):
        try:
            self.compiler.compile()
        except Exception as e:
            logging.error(f"Compilation failed: {str(e)}")


class MeltanoBackgroundCompiler:
    def __init__(self, project: Project, compiler: ProjectCompiler = None):
        self.project = project
        self.compiler = compiler or ProjectCompiler(project)
        self.observer = self.setup_observer()

    @property
    def model_dir(self):
        return self.project.root_dir("model")

    def setup_observer(self):
        event_handler = CompileEventHandler(self.compiler)
        observer = Observer()
        observer.schedule(event_handler, str(self.model_dir), recursive=True)

        return observer

    def start(self):
        try:
            self.observer.start()
            logging.info(f"Auto-compiling models in '{self.model_dir}'")
        except OSError:
            # most probably INotify being full
            logging.warn(f"Model auto-compilation is disabled: INotify limit reached.")

    def stop(self):
        self.observer.stop()


class UIAvailableWorker(threading.Thread):
    def __init__(self, url, open_browser=False):
        super().__init__()

        self._terminate = False
        self.url = url
        self.open_browser = open_browser

    def run(self):
        while not self._terminate:
            try:
                response = requests.get(self.url)
                if response.status_code == 200:
                    print(f"{Fore.GREEN}Meltano is available at {self.url}{Fore.RESET}")
                    if self.open_browser:
                        webbrowser.open(self.url)
                    self._terminate = True

            except:
                pass

            time.sleep(2)

    def stop(self):
        self._terminate = True


class AirflowWorker(threading.Thread):
    def __init__(self, project: Project):
        super().__init__()

        self.project = project
        self.installed = threading.Event()
        self._plugin = None
        self._webserver = None
        self._scheduler = None

        try:
            if ConfigService(self.project).find_plugin("airflow"):
                self.installed.set()
        except:
            pass

    def kill_stale_workers(self):
        stale_workers = []
        workers_pid = map(self.pid_path, ("webserver", "scheduler"))

        for f in workers_pid:
            try:
                pid = int(f.open().read())
                stale_workers.append(psutil.Process(pid))
            except psutil.NoSuchProcess:
                f.unlink()
            except FileNotFoundError:
                pass

        def on_terminate(proc):
            logging.info(f"Process {proc} ended with exit code {proc.returncode}")

        for p in stale_workers:
            logging.debug(f"Process {p} is stale, terminating it.")
            p.terminate()

        gone, alive = psutil.wait_procs(stale_workers, timeout=5, callback=on_terminate)

        # kill the rest
        for p in alive:
            p.kill()

    def start_all(self):
        logs_dir = self.project.run_dir("airflow", "logs")
        invoker = invoker_factory(db.session, self.project, self._plugin)

        with logs_dir.joinpath("webserver.log").open(
            "w"
        ) as webserver, logs_dir.joinpath("scheduler.log").open(
            "w"
        ) as scheduler, self.pid_path(
            "webserver"
        ).open(
            "w"
        ) as webserver_pid, self.pid_path(
            "scheduler"
        ).open(
            "w"
        ) as scheduler_pid:
            self._webserver = invoker.invoke("webserver", "-w", "1", stdout=webserver)
            self._scheduler = invoker.invoke("scheduler", stdout=scheduler)

            webserver_pid.write(str(self._webserver.pid))
            scheduler_pid.write(str(self._scheduler.pid))

        # Time padding for server initialization so UI iframe displays as expected
        # (iteration potential on approach but following UIAvailableWorker sleep approach)
        time.sleep(2)

    def pid_path(self, name):
        return self.project.run_dir("airflow", f"{name}.pid")

    def run(self):
        self.installed.wait()
        self._plugin = ConfigService(self.project).find_plugin("airflow")
        self.kill_stale_workers()
        self.start_all()

    def stop(self):
        self.kill_stale_workers()



#!/usr/bin/env python3
# TODO: convert this to alembic http://alembic.zzzcomputing.com/en/latest/
import backoff
from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError

from meltano.api.models import *
from meltano.core.job import *
from meltano.core.plugin.setting import *
from meltano.core.db import SystemMetadata


# to get the desired UX we should wait for the db to be available
@backoff.on_exception(backoff.expo, OperationalError, max_time=60)
def create_db():
    engine = create_engine("sqlite:///meltano.db")
    SystemMetadata.create_all(engine)


if __name__ == "__main__":
    create_db()



import subprocess
import json
import logging
import asyncio
import os
from pathlib import Path
from datetime import datetime
from enum import IntFlag

from . import Runner
from meltano.core.db import project_engine
from meltano.core.job import Job, JobFinder
from meltano.core.project import Project
from meltano.core.plugin_invoker import invoker_factory, PluginInvoker
from meltano.core.config_service import ConfigService
from meltano.core.plugin.singer import SingerTap, SingerTarget, PluginType
from meltano.core.utils import file_has_data


class SingerPayload(IntFlag):
    STATE = 1


class SingerRunner(Runner):
    def __init__(
        self, project: Project, job_id, config_service: ConfigService = None, **config
    ):
        self.project = project
        self.job_id = job_id
        self.config_service = config_service or ConfigService(project)
        self.config = config

        self.job = Job(job_id=self.job_id)
        self.run_dir = Path(config.get("run_dir", "/run/singer"))
        self.tap_config_dir = Path(config.get("tap_config_dir", "/etc/singer/tap"))
        self.target_config_dir = Path(
            config.get("target_config_dir", "/etc/singer/target")
        )

        _, self._session_cls = project_engine(project)

    @property
    def database(self):
        return self.config.get("database", "default")

    def stop(self, process, **wait_args):
        while True:
            try:
                code = process.wait(**wait_args)
                logging.debug(f"{process} exited with {code}")
                return code
            except subprocess.TimeoutExpired:
                process.kill()
                logging.error(f"{process} was killed.")

    def prepare(self, tap: PluginInvoker, target: PluginInvoker):
        tap.prepare()
        target.prepare()

    async def invoke(self, tap: PluginInvoker, target: PluginInvoker):
        try:
            # use standard pipes here because we
            # don't need async processing between the
            # tap and target.
            target_in, tap_out = os.pipe()

            p_target, p_tap = None, None
            p_target = await target.invoke_async(
                stdin=target_in, stdout=asyncio.subprocess.PIPE  # state log
            )
            os.close(target_in)

            p_tap = await tap.invoke_async(stdout=tap_out)
            os.close(tap_out)
        except Exception as err:
            if p_tap:
                p_tap.kill()
            if p_target:
                p_target.kill()
            raise Exception(f"Cannot start tap or target: {err}")

        # receive the target stdout and update the current job
        # for each lines
        await asyncio.wait(
            [self.bookmark(p_target.stdout), p_target.wait(), p_tap.wait()],
            return_when=asyncio.FIRST_COMPLETED,
        )

        # at this point, something already stopped, the other component
        # should die soon because of a SIGPIPE
        target_code = await p_target.wait()
        tap_code = await p_tap.wait()

        if any((tap_code, target_code)):
            raise Exception(
                f"Subprocesses didn't exit cleanly: tap({tap_code}), target({target_code})"
            )

    def restore_bookmark(self, session, tap: PluginInvoker):
        # the `state.json` is stored in the database
        finder = JobFinder(self.job_id)
        state_job = finder.latest_with_payload(session, flags=SingerPayload.STATE)

        if state_job and "singer_state" in state_job.payload:
            logging.info(f"Found state from {state_job.started_at}.")
            with tap.files["state"].open("w+") as state:
                json.dump(state_job.payload["singer_state"], state)
        else:
            logging.warning("No state was found, complete import.")

    def bookmark_state(self, new_state: str):
        try:
            new_state = json.loads(new_state)

            self.job.payload["singer_state"] = new_state
            self.job.payload_flags |= SingerPayload.STATE
            self.job.ended_at = datetime.utcnow()
            logging.info(f"Incremental state has been updated at {self.job.ended_at}.")
            logging.debug(f"Incremental state: {new_state}")
        except Exception as err:
            logging.warning(
                "Received state is invalid, incremental state has not been updated"
            )

    async def bookmark(self, target_stream):
        while not target_stream.at_eof():
            last_state = await target_stream.readline()
            if not last_state:
                continue

            self.bookmark_state(last_state)

    def dry_run(self, extractor: SingerTap, loader: SingerTarget):
        tap_exec = extractor.exec_path()
        target_exec = loader.exec_path()

        logging.info("Dry run:")
        logging.info(f"\textractor: {extractor.name} at '{tap_exec}'")
        logging.info(f"\tloader: {extractor.name} at '{target_exec}'")

    def run(self, extractor: str, loader: str, dry_run=False):
        tap = self.config_service.find_plugin(
            extractor, plugin_type=PluginType.EXTRACTORS
        )
        target = self.config_service.find_plugin(loader, plugin_type=PluginType.LOADERS)

        try:
            session = self._session_cls()
            extractor = invoker_factory(session, self.project, tap)
            loader = invoker_factory(session, self.project, target)

            self.prepare(extractor, loader)

            if dry_run:
                return self.dry_run(tap, target)

            with self.job.run(session):
                self.restore_bookmark(session, extractor)
                loop = asyncio.get_event_loop()
                loop.run_until_complete(self.invoke(extractor, loader))
        finally:
            session.close()



import logging
import os
from sqlalchemy import create_engine

import meltano.api.config as _config
from meltano.core.project import Project
from meltano.core.migration_service import MigrationService
from meltano.api.workers import MeltanoBackgroundCompiler, AirflowWorker


_port = os.getenv("MELTANO_API_PORT", "5000")
_host = os.getenv("MELTANO_API_HOSTNAME", "0.0.0.0")

bind = [f"{_host}:{_port}"]
project = Project.find()

_workers = []
_workers.append(MeltanoBackgroundCompiler(project))
try:
    _workers.append(AirflowWorker(project))
except:
    logging.info("Airflow is not installed.")


def when_ready(server):
    engine = create_engine(_config.SQLALCHEMY_DATABASE_URI)
    MigrationService(engine).upgrade()
    engine.dispose()

    for worker in _workers:
        worker.start()


def on_exit(server):
    for worker in _workers:
        worker.stop()



import subprocess
import logging

from .config_service import ConfigService
from .db import project_engine
from .plugin import PluginType
from .plugin_invoker import invoker_factory
from .venv_service import VenvService


class DbtService:
    def __init__(self, project):
        self.project = project
        self._plugin = ConfigService(project).find_plugin(
            "dbt", PluginType.TRANSFORMERS
        )
        self.profile_dir = f"{self.project.root}/transform/profile/"
        _, self._Session = project_engine(project)

    @property
    def exec_path(self):
        return self.venv_service.exec_path("dbt", namespace=PluginType.TRANSFORMERS)

    def compile(self, models=None, **kwargs):
        session = self._Session()
        invoker = invoker_factory(session, self.project, self._plugin)

        try:
            params = ["--profiles-dir", self.profile_dir, "--profile", "meltano"]
            if models:
                # Always include the my_meltano_project model
                all_models = f"{models} my_meltano_project"
                params.extend(["--models", all_models])

            handle = invoker.invoke("compile", *params, **kwargs)
            handle.wait()

            return handle
        finally:
            session.close()

    def deps(self):
        session = self._Session()
        invoker = invoker_factory(session, self.project, self._plugin)
        try:
            handle = invoker.invoke("deps")
            handle.wait()

            return handle
        finally:
            session.close()

    def run(self, models=None, **kwargs):
        session = self._Session()
        invoker = invoker_factory(session, self.project, self._plugin)

        try:
            params = ["--profiles-dir", self.profile_dir, "--profile", "meltano"]
            if models:
                # Always include the my_meltano_project model
                all_models = f"{models} my_meltano_project"
                params.extend(["--models", all_models])

            handle = invoker.invoke("run", *params, **kwargs)
            handle.wait()

            return handle
        finally:
            session.close()



import datetime
import logging
import logging.handlers
import os
import atexit
from flask import Flask, request, g
from flask_login import current_user
from flask_cors import CORS
from importlib import reload
from urllib.parse import urlsplit

from meltano.core.project import Project
from meltano.core.plugin.error import PluginMissingError
from meltano.core.plugin.settings_service import (
    PluginSettingsService,
    PluginSettingMissingError,
)
from meltano.core.config_service import ConfigService
from meltano.core.compiler.project_compiler import ProjectCompiler
from .workers import MeltanoBackgroundCompiler, UIAvailableWorker, AirflowWorker


logger = logging.getLogger(__name__)
airflow_context = {"worker": None}


def create_app(config={}):
    project = Project.find()

    app = Flask(
        __name__, instance_path=str(project.root), instance_relative_config=True
    )

    app.config.from_object("meltano.api.config")
    app.config.from_pyfile("ui.cfg", silent=True)
    app.config.update(**config)

    # the database should be instance_relative if we are using `sqlite`
    scheme, netloc, path, *parts = urlsplit(app.config["SQLALCHEMY_DATABASE_URI"])
    if scheme == "sqlite" and path:
        app.config["SQLALCHEMY_DATABASE_URI"] = (
            scheme + ":///" + app.instance_path + path
        )

    # Initial compilation
    compiler = ProjectCompiler(project)
    try:
        compiler.compile()
    except Exception as e:
        pass

    # Logging
    file_handler = logging.handlers.RotatingFileHandler(
        str(project.run_dir("meltano-ui.log")), backupCount=3
    )
    stdout_handler = logging.StreamHandler()

    logger.setLevel(logging.DEBUG)
    logger.addHandler(file_handler)
    logger.addHandler(stdout_handler)

    now = str(datetime.datetime.utcnow().strftime("%b %d %Y %I:%M:%S:%f"))
    logger.warning(f"Meltano UI started at: {now}")

    # Extensions
    security_options = {}

    from .models import db
    from .mail import mail
    from .executor import setup_executor
    from .security import security, users, setup_security
    from .security.oauth import setup_oauth
    from .json import setup_json

    db.init_app(app)
    mail.init_app(app)
    setup_executor(app, project)
    setup_security(app, project)
    setup_oauth(app)
    setup_json(app)
    CORS(app, origins="*")

    from .controllers.root import root
    from .controllers.dashboards import dashboardsBP
    from .controllers.reports import reportsBP
    from .controllers.repos import reposBP
    from .controllers.settings import settingsBP
    from .controllers.sql import sqlBP
    from .controllers.orchestrations import orchestrationsBP
    from .controllers.plugins import pluginsBP

    app.register_blueprint(root)
    app.register_blueprint(dashboardsBP)
    app.register_blueprint(reportsBP)
    app.register_blueprint(reposBP)
    app.register_blueprint(settingsBP)
    app.register_blueprint(sqlBP)
    app.register_blueprint(orchestrationsBP)
    app.register_blueprint(pluginsBP)

    if app.config["PROFILE"]:
        from .profiler import init

        init(app)

    @app.before_request
    def setup_airflow_context():
        g.airflow_worker = airflow_context["worker"]

    @app.before_request
    def setup_js_context():
        appUrl = urlsplit(request.host_url)
        g.jsContext = {"appUrl": appUrl.geturl()[:-1]}

        try:
            airflow = ConfigService(project).find_plugin("airflow")
            settings = PluginSettingsService(db.session, project)
            airflow_port, _ = settings.get_value(airflow, "webserver.web_server_port")
            g.jsContext["airflowUrl"] = appUrl._replace(
                netloc=f"{appUrl.hostname}:{airflow_port}"
            ).geturl()[:-1]
        except (PluginMissingError, PluginSettingMissingError):
            pass

    @app.after_request
    def after_request(res):
        request_message = f"[{request.url}]"

        if request.method != "OPTIONS":
            request_message += f" as {current_user}"

        logger.info(request_message)
        return res

    return app


def start(project, **kwargs):
    """Start Meltano UI as a single-threaded web server."""

    app_config = kwargs.pop("app_config", {})
    app = create_app(app_config)
    from .security.identity import create_dev_user

    with app.app_context():
        # TODO: alembic migration
        create_dev_user()

    # ensure we only start the workers on the via the main thread
    # this will make sure we don't start everything twice
    # when code reload is enabled
    if not app.debug or os.environ.get("WERKZEUG_RUN_MAIN") == "true":
        cleanup = start_workers(app, project, **kwargs)
        atexit.register(cleanup)

    app.run(**kwargs)


def start_workers(app, project, **kwargs):
    port = kwargs.pop("port", 5000)
    workers = []

    if not app.config["AIRFLOW_DISABLED"]:
        airflow_context["worker"] = AirflowWorker(project)
        workers.append(airflow_context["worker"])

    workers.append(MeltanoBackgroundCompiler(project))
    workers.append(
        # the web server should always be accessible from `localhost`
        UIAvailableWorker(f"http://localhost:{port}", open_browser=not app.debug)
    )

    # cleanup callback
    def stop_all():
        logger.info("Stopping all background workers...")
        for worker in workers:
            worker.stop()

    # start all workers
    for worker in workers:
        worker.start()

    return stop_all



import os
import click
from pathlib import Path
from sqlalchemy import create_engine
from alembic.config import Config
from alembic.script import ScriptDirectory
from alembic import command

from meltano.core.migration_service import MigrationService
from meltano.core.db import project_engine
from . import cli
from .params import project, db_options


@cli.command()
@db_options
@click.pass_context
def upgrade(ctx, engine_uri):
    engine = create_engine(engine_uri)

    migration_service = MigrationService(engine)
    migration_service.upgrade()



import os
from airflow import DAG
from airflow.operators.bash_operator import BashOperator
from datetime import datetime, time, timedelta, MINYEAR

from meltano.core.schedule_service import ScheduleService
from meltano.core.project import Project
from meltano.core.utils import coerce_datetime


project = Project.find()

# I really don't like sending None here because it
# totally breaks the encapsulation. We should decouple
# `Session` from the service.
schedule_service = ScheduleService(None, project)

default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "email_on_failure": False,
    "email_on_retry": False,
    "retries": 1,
    "retry_delay": timedelta(minutes=5),
    "concurrency": 1,
}

for schedule in schedule_service.schedules():
    catchup = False
    args = default_args.copy()

    if schedule.start_date:
        args["start_date"] = coerce_datetime(schedule.start_date)
        catchup = True

    dag_id = f"meltano_{schedule.name}"
    dag = DAG(
        dag_id, default_args=args, schedule_interval=schedule.interval, catchup=catchup
    )

    elt = BashOperator(
        task_id="extract_load",
        bash_command=f"echo $PATH; echo $VIRTUAL_ENV; cd {str(project.root)}; .meltano/run/bin elt {schedule.extractor} {schedule.loader} --transform={schedule.transform}",
        dag=dag,
        env={
            # inherit the current env
            **os.environ,
            **schedule.env,
        },
    )

    # register the dag
    globals()[dag_id] = dag



import logging

from flask import Blueprint, request, url_for, jsonify, make_response, Response

from meltano.core.job import JobFinder, State
from meltano.core.plugin import PluginRef
from meltano.core.plugin.error import PluginExecutionError
from meltano.core.plugin.settings_service import (
    PluginSettingsService,
    PluginSettingValueSource,
)
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.project import Project
from meltano.core.project_add_service import ProjectAddService
from meltano.core.config_service import ConfigService
from meltano.core.schedule_service import ScheduleService, ScheduleAlreadyExistsError
from meltano.core.select_service import SelectService
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.utils import flatten, iso8601_datetime, slugify
from meltano.cli.add import extractor
from meltano.api.models import db
from meltano.api.json import freeze_keys

from meltano.api.executor import run_elt


orchestrationsBP = Blueprint(
    "orchestrations", __name__, url_prefix="/api/v1/orchestrations"
)


@orchestrationsBP.errorhandler(ScheduleAlreadyExistsError)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"A schedule with the name '{ex.schedule.name}' already exists. Try renaming the schedule.",
            }
        ),
        409,
    )


@orchestrationsBP.route("/connection_names", methods=["GET"])
def connection_names():
    settings = Settings.query.first()
    if not settings:
        settings = Settings()
    connections = [
        c["name"] for c in settings.serializable()["settings"]["connections"]
    ]
    return jsonify(connections)


@orchestrationsBP.route("/job/state", methods=["POST"])
def job_state() -> Response:
    """
    endpoint for getting the status of N jobs
    """
    project = Project.find()
    poll_payload = request.get_json()
    job_ids = poll_payload["job_ids"]

    jobs = []
    for job_id in job_ids:
        finder = JobFinder(job_id)
        state_job = finder.latest(db.session)
        # Validate existence first as a job may not be queued yet as a result of
        # another prerequisite async process (dbt installation for example)
        if state_job:
            is_complete = state_job.state == State.SUCCESS
            jobs.append({"job_id": job_id, "is_complete": is_complete})

    return jsonify({"jobs": jobs})


@orchestrationsBP.route("/run", methods=["POST"])
def run():
    project = Project.find()
    schedule_payload = request.get_json()
    job_id = run_elt(project, schedule_payload)

    return jsonify({"job_id": job_id}), 202


@orchestrationsBP.route("/get/configuration", methods=["POST"])
def get_plugin_configuration() -> Response:
    """
    endpoint for getting a plugin's configuration
    """
    project = Project.find()
    payload = request.get_json()
    plugin = PluginRef(payload["type"], payload["name"])
    settings = PluginSettingsService(db.session, project)

    config = flatten(settings.as_config(plugin), reducer="dot")

    return jsonify(
        {
            # freeze the keys because they are used for lookups
            "config": freeze_keys(config),
            "settings": settings.get_definition(plugin).settings,
        }
    )


@orchestrationsBP.route("/save/configuration", methods=["POST"])
def save_plugin_configuration() -> Response:
    """
    endpoint for persisting a plugin configuration
    """
    project = Project.find()
    incoming = request.get_json()
    plugin = PluginRef(incoming["type"], incoming["name"])
    config = incoming["config"]

    settings = PluginSettingsService(db.session, project)
    for name, value in config.items():
        if value == "":
            settings.unset(plugin, name)
        else:
            settings.set(plugin, name, value)

    return jsonify(settings.as_config(plugin))


@orchestrationsBP.route("/select-entities", methods=["POST"])
def selectEntities() -> Response:
    """
    endpoint that performs selection of the user selected entities and attributes
    """
    project = Project.find()
    incoming = request.get_json()
    extractor_name = incoming["extractor_name"]
    entity_groups = incoming["entity_groups"]
    select_service = SelectService(project, extractor_name)

    for entity_group in entity_groups:
        group_is_selected = "selected" in entity_group

        for attribute in entity_group["attributes"]:
            if group_is_selected or "selected" in attribute:
                entities_filter = entity_group["name"]
                attributes_filter = attribute["name"]
                select_service.select(entities_filter, attributes_filter)

    return jsonify("winning")


@orchestrationsBP.route("/entities/<extractor_name>", methods=["POST"])
def entities(extractor_name: str) -> Response:
    """
    endpoint that returns the entities associated with a particular extractor
    """
    project = Project.find()
    select_service = SelectService(project, extractor_name)

    entity_groups = []
    try:
        list_all = select_service.get_extractor_entities()

        for stream, prop in (
            (stream, prop)
            for stream in list_all.streams
            for prop in list_all.properties[stream.key]
        ):
            match = next(
                (
                    entityGroup
                    for entityGroup in entity_groups
                    if entityGroup["name"] == stream.key
                ),
                None,
            )
            if match:
                match["attributes"].append({"name": prop.key})
            else:
                entity_groups.append(
                    {"name": stream.key, "attributes": [{"name": prop.key}]}
                )

        entity_groups = sorted(entity_groups, key=lambda k: k["name"])
        for entityGroup in entity_groups:
            entityGroup["attributes"] = sorted(
                entityGroup["attributes"], key=lambda k: k["name"]
            )
    except PluginExecutionError as e:
        logging.warning(str(e))

    return jsonify({"extractor_name": extractor_name, "entity_groups": entity_groups})


@orchestrationsBP.route("/get/pipeline_schedules", methods=["GET"])
def get_pipeline_schedules():
    """
    endpoint for getting the pipeline schedules
    """
    project = Project.find()
    schedule_service = ScheduleService(db.session, project)
    schedules = [s._asdict() for s in schedule_service.schedules()]
    for schedule in schedules:
        finder = JobFinder(f"job_{schedule['name']}")
        state_job = finder.latest(db.session)
        is_running = state_job.state is State.RUNNING if state_job else False
        schedule["is_running"] = is_running

    return jsonify(schedules)


@orchestrationsBP.route("/save/pipeline_schedule", methods=["POST"])
def save_pipeline_schedule() -> Response:
    """
    endpoint for persisting a pipeline schedule
    """
    incoming = request.get_json()
    # Airflow requires alphanumeric characters, dashes, dots and underscores exclusively
    name = slugify(incoming["name"])
    extractor = incoming["extractor"]
    loader = incoming["loader"]
    transform = incoming["transform"]
    interval = incoming["interval"]
    start_date = incoming["start_date"]

    project = Project.find()
    schedule_service = ScheduleService(db.session, project)

    try:
        schedule = schedule_service.add(
            name,
            extractor,
            loader,
            transform,
            interval,
            start_date=iso8601_datetime(start_date),
        )
        return jsonify(schedule._asdict()), 201
    except ScheduleAlreadyExistsError as e:
        raise ScheduleAlreadyExistsError(e.schedule)



import logging
import subprocess
from io import StringIO

from . import Runner
from meltano.core.project import Project
from meltano.core.plugin import PluginType
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.dbt_service import DbtService
from meltano.core.config_service import ConfigService
from meltano.core.db import project_engine


class DbtRunner(Runner):
    def __init__(
        self,
        project: Project,
        config_service: ConfigService = None,
        dbt_service: DbtService = None,
        plugin_settings_service=None,
    ):
        self.project = project
        self.config_service = config_service or ConfigService(project)
        self.dbt_service = dbt_service or DbtService(project)

    def run(self, extractor: str, loader: str, dry_run=False, models=None):
        extractor = self.config_service.find_plugin(
            extractor, plugin_type=PluginType.EXTRACTORS
        )
        loader = self.config_service.find_plugin(loader, plugin_type=PluginType.LOADERS)

        # we should probably refactor this part to have an ELTContext object already
        # filled with the each plugins' configuration so we don't have to query
        # multiple times for the same data.
        _, Session = project_engine(self.project)

        session = Session()
        try:
            settings = PluginSettingsService(session, self.project)

            # send the elt_context as ENV variables
            env = {**settings.as_env(extractor), **settings.as_env(loader)}
            process = self.dbt_service.deps()

            if process.returncode:
                raise Exception(
                    f"dbt deps didn't exit cleanly. Exit code: {process.returncode}"
                )

            if models is not None:
                models = models.replace("-", "_")

            if dry_run:
                process = self.dbt_service.compile(models, env=env)
                if process.returncode:
                    raise Exception(
                        f"dbt compile didn't exit cleanly. Exit code: {process.returncode}"
                    )
            else:
                process = self.dbt_service.run(models, env=env)
                if process.returncode:
                    raise Exception(
                        f"dbt run didn't exit cleanly. Exit code: {process.returncode}"
                    )
        except Exception as e:
            logging.warning("Could not inject environment to dbt.")
            logging.debug(f"Could not hydrate ENV from the EltContext: {str(e)}")
            raise e
